/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	UNICNT:					*/
/*						*/
/************************************************/
#include        <sys/mount.h>
#include        <sys/syscall.h>
#include        <sys/time.h>
#include        <sys/wait.h>
#include        <sys/stat.h>
#include	<time.h>
#include	<signal.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<unistd.h>

#include	"lowtyp.h"
#include	"dbgmem.h"
#include	"subapl.h"
#include	"unicnt.h"

/*container filesystem				*/
#define	HOMEFS	"HOMEFS"	/*container HOME*/
#define	ROOTFS	"rootfs"	/*container root*/
#define	OLDROOT	".oldroot"	/*for pivot_root*/
#define	CLONPID	"first.pid"	/*cont process 1*/
#define	MTAB	"/etc/mtab"	/*mtab file	*/

#define	DPROC	"/proc"		/*kernel /proc	*/
#define	DSYS	"/sys"		/*kernel /sys	*/
#define	DDEV	"/dev"		/*system /dev	*/
#define	DEVFS	"devtmpfs"	/*/dev systype	*/
/*systemd sugget to have /proc/ and /sys to be	*/
/*mounted read only (MS_RDONLY).		*/
#define	MNTMODE	0		/*Mounting mode	*/

/*

*/
/************************************************/
/*						*/
/*	Defining pivot_root system call		*/
/*						*/
/************************************************/
static int pivot_root(const char * new_root,const char * put_old)

{
#ifndef __NR_pivot_root
#pragma message ("the pivot_root syscall is not available within \"sys/syscall.h\"")
#pragma message ("-> The pivot_root system call will generate an alert")
(void) apl_alert(0,"%s need the pivot_root system call!!!.",appname);
errno=ENOSYS;
return -1;
#else
return(syscall(__NR_pivot_root, new_root, put_old));
#endif
}
/*

*/
/************************************************/
/*						*/
/*	Returning the container path.		*/
/*	contpath array is dynamically assigned	*/
/*	must be freed when not needed anymore.	*/
/*						*/
/************************************************/
static char *getcontpath(char *contname)

{
char *homefs;
char *contpath;
char *vzpath;


homefs=getenv(HOMEFS);
contpath=(char *)0;
if (homefs==(char *)0) {
  vzpath=apl_appdir(d_vzgot);
  (void) asprintf(&contpath,"%s/%s",vzpath,contname);
  (void) free(vzpath);
  }
else
  (void) asprintf(&contpath,"%s",homefs);
return contpath;
}
/*

*/
/************************************************/
/*						*/
/*	Check is the container is here with	*/
/*	all needed parts.			*/
/*						*/
/************************************************/
int cnt_iscontgood(char *contname)

{
#define	OPEP	"unicnt.c:cnt_iscontgood,"

int isgood;
char *contpath;
int phase;
int proceed;

isgood=true;
phase=0;
proceed=true;
contpath=getcontpath(contname);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is container existing?*/
      if (contpath!=(char *)0) {/*always.	*/
	char *rootpath;	

	(void) asprintf(&rootpath,"%s/%s",contpath,ROOTFS);
        if (apl_isdir(rootpath)<0) {
          (void) apl_alert(0,"%s rootfs directory <%s> missing "
			     "within container <%s> directory",
			     OPEP,rootpath,contname);
	  isgood=false;
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(rootpath);
	}
      break;
    case 1	:	/*check/make oldroot dir*/
      if (contpath!=(char *)0) {/*always.	*/
	char *oldpath;

	(void) asprintf(&oldpath,"%s/%s",contpath,OLDROOT);
        if (apl_isdir(oldpath)<0) {
	  if (mkdir(oldpath,0700)<0) {
            (void) apl_alert(0,"%s, unable to make %s directory "
			       "for container <%s> (error=%s)",
			       OPEP,OLDROOT,contname,strerror(errno));
	    isgood=false;
	    phase=999;	/*trouble trouble	*/
	    }
	  }
	(void) free(oldpath);
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(contpath);
return isgood;
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to pivot the container root	*/
/*	file system.				*/
/*						*/
/************************************************/
int cnt_pivot(int foreground,char *contname)

{
#define	OPEP	"unicnt.c:cnt_pivot,"

int done;
char *rootfs;
char *contpath;
int phase;
int proceed;

done=false;
rootfs=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&rootfs,"%s/%s",contpath,ROOTFS);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*binding container file*/
      if (mount(rootfs,rootfs,0,MS_BIND|MS_RDONLY,0)<0) {
        (void) apl_alert(0,"%s unable to bind container <%s> directory "
			   "(error=%s)",
			   OPEP,rootfs,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      //(void) apl_alert(0,"%s rootfs=<%s> -> phase='%d'","JMPDBG",rootfs,phase);
      break;
    case 1	:	/*moving to rootfs	*/
      if (chdir(contpath)<0) {
        (void) apl_alert(0,"%s unable to reach container <%s> "
			   "(directory=<%s>,  error=%s)",
			   OPEP,contname,contpath,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*moving to rootfs	*/
      //(void) apl_alert(0,"%s PWD=<%s>","JMPDBG",get_current_dir_name());
      if (pivot_root(".",OLDROOT)<0) {
        (void) apl_alert(0,"%s unable to pivot container <%s> "
			   "'.'=<%s> oldroot=<%s> (error=%s)",
			   OPEP,contname,contpath,OLDROOT,strerror(errno));
	phase=999;	/*trouble trouble	*/
	(void) exit(1);	//JMPDBG
	}
      break;
    case 3	:	/*umounting previous fs	*/
      if (umount2(OLDROOT,MNT_DETACH)<0) {
        (void) apl_alert(0,"%s unable to unmount container <%s> "
			   "oldroot (error=%s)",
			   OPEP,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/* chrooting to rootfs	*/
      if (rmdir(OLDROOT)<0) 
        (void) apl_alert(0,"%s unable to remove <%s> directory "
			   "from container <%s>, (error=%s)",
			   OPEP,OLDROOT,contname,strerror(errno));
      if (chdir(ROOTFS)<0) {
        (void) apl_alert(0,"%s unable to reach container <%s> "
			   "rootfs (error=%s)",
			   OPEP,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 5	:	/* everything fine	*/
      if (chroot(".")<0) {
        (void) apl_alert(0,"%s unable to rootfs container <%s> (error=%s)",
			   OPEP,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 6	:	/*remounting /root ?	*/
      break;
    case 7	:	/*checking /proc dir	*/
      if (access(DPROC,F_OK)<0) {
        (void) apl_alert(0,"%s problem with directory <%s> (error=%s)",
			   OPEP,DPROC,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 8	:	/*checking /sys dir	*/
      if (access(DSYS,F_OK)<0) {
        (void) apl_alert(0,"%s problem with directory <%s> (error=%s)",
			    OPEP,DSYS,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 9	:	/*mounting proc dir	*/
      if (mount("proc",DPROC,"proc",MNTMODE,0)<0) {
        (void) apl_alert(0,"%s Unable to mount directory <%s> (error=%s)",
			    OPEP,DPROC,strerror(errno));
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 10	:	/*mounting sys dir	*/
      if (mount("sysfs",DSYS,"sysfs",MNTMODE,0)<0) {
        (void) apl_alert(0,"%s Unable to mount directory <%s> (error=%s)",
			    OPEP,DSYS,strerror(errno));
	(void) umount(DPROC);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 11	:	/*mounting dev RD_ONLY?	*/
      break;
    case 12	:	/* everything fine	*/
      (void) fprintf(stdout,"%s chroot done for container <%s>\n",
			    apl_ascsystime(time((time_t *)0)),contname);
      (void) fflush(stdout);
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(contpath);
(void) free(rootfs);
return done;
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set a file with the clone	*/
/*	ip number.				*/
/*						*/
/************************************************/
int cnt_setclonepid(char *contname,pid_t cpid)

{
int done;
FILE *fichier;
char *filename;
char *contpath;
int phase;
int proceed;

done=false;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((fichier=fopen(filename,"w"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to open pidfile <%s> (error=<%s>)",
			    appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*writing  PID in it	*/
      if (fprintf(fichier,"%d\n",cpid)<0) {
	(void) apl_alert(0,"%s, Unable to write pidfile in container <%s> (error=<%s>)",
			    appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*everything fine	*/
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get the clone pid stored	*/
/*	within the clonpid file			*/
/*						*/
/************************************************/
pid_t cnt_getclonepid(char *contname)

{
pid_t clonepid;
FILE *fichier;
char *filename;
char *contpath;

clonepid=(pid_t)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  char strloc[80];

  (void) fgets(strloc,sizeof(strloc)-1,fichier);
  (void) sscanf(strloc,"%d",&clonepid);
  (void) fclose(fichier);
  }
(void) free(filename);
(void) free(contpath);
return clonepid;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to remove the file with the	*/
/*	clone pid.				*/
/*						*/
/************************************************/
int cnt_rmclonepid(char *contname)

{
int done;
char *filename;
char *contpath;

done=true;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if (unlink(filename)<0) {
  (void) apl_alert(0,"%s, Unable to remove pidfile in container <%s> (error=<%s>)",
		     appname,contname,strerror(errno));
  done=false;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to initialize the HOST side	*/
/*						*/
/************************************************/
int cnt_initscript(const char *scriptname,const char *fmt,...)

{
#define	BSIZE	2048

int status;
va_list args;
FILE *canal;
char *varlibpath;
char *parmlst;
char *cmd;
int phase;
int proceed;

status=0;
va_start(args,fmt);
canal=(FILE *)0;
varlibpath=apl_appdir(d_varlib);
(void) vasprintf(&parmlst,fmt,args);
(void) asprintf(&cmd,"%s/shell/%s %s",varlibpath,scriptname,parmlst);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening pip channel	*/
      if ((canal=popen(cmd,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to pipe cmd <%s> (error=<%s>)",
			    appname,cmd,strerror(errno));
	status=1;
	phase=999;	/*trouble trouble	*/	
	}
      break;
    case 1	:	/*do we have feed back	*/
      if (canal!=(FILE *)0) {	/*always	*/
	char buffer[BSIZE];

	while (fgets(buffer,BSIZE,canal)!=(char *)0) {
	  (void) apl_cleanstring(buffer);
	  if ((debug>0)&&(verbose==true))
	    (void) apl_alert(0,"%s, pipe say: <%s>",appname,buffer);
	  else {
	    (void) fprintf(stdout,"%s\n",buffer);
	    (void) fflush(stdout);
	    }
	  }
	}
      break;
    case 2	:	/*closing pip		*/
      status=pclose(canal);/*get pipe status	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cmd);
(void) free(parmlst);
(void) free(varlibpath);
va_end(args);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to inject a command within	*/
/*	the container.				*/
/*						*/
/************************************************/
int cnt_injectcmd(const char *scriptname,char *contname,int contpid,char *params)

{
#define	OPEP	"unicnt.c:cnt_injectcmd,"
#define	FDS	3
#define	CMDSEQ	"%s/shell/%s %s %d %s"

int status;
char *varlibpath;
char *cmd;
int phase;
int proceed;

status=0;
varlibpath=apl_appdir(d_varlib);
(void) asprintf(&cmd,CMDSEQ,varlibpath,scriptname,contname,contpid,params);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	//protecting already std devices?
      break;
    case 1	:	//executing the command
      status=system(cmd);
      break;
    case 2	:	//putting back devices
      break;
    default	:	//SAFE Guard
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cmd);
(void) free(varlibpath);
return status;
#undef OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	architecture available in the 'arch' 	*/
/*	file.					*/
/*						*/
/************************************************/
char *cnt_getarch(char *contname)

{
static char *availarch[]={
	"i386","i686","x86_64",(char *)0
	};

char *arch;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

arch=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"arch");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      if (strlen(buffer)>0) { 	/*always?	*/
	int i;

        for (i=0;availarch[i]!=(char *)0;i++) {
	  if (strcmp(availarch[i],buffer)==0) {
	    arch=strdup(availarch[i]);
	    proceed=false;	/*found arch	*/
	    break;
	    }
	  }
	}
      break;
    case 3	:	/*scanning line		*/
      (void) apl_alert(0,"%s file <%s>, arch <%s> unknown, setting default",
			   appname,filename,buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (arch==(char *)0) {
  arch=strdup(availarch[0]);
  }
return arch;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	distribution type available in the  	*/
/*	'dist' file.				*/
/*						*/
/************************************************/
char *cnt_getdist(char *contname)

{
char *dist;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

dist=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"dist");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      dist=strdup(buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (dist==(char *)0) {
  dist=strdup("unknown");
  }
return dist;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to assign a new STDOUT/STDERR	*/
/*	and close STDIN. Done ONLY if the	*/
/*	process is in background mode.		*/
/*						*/
/************************************************/
void cnt_setstdio(int foreground,char *contname,char *outname)

{
#define	OPEP	"unicnt.c:cnt_setstdio,"

if (foreground==false) {
  int newstdout;
  int newstderr;
  int phase;
  int proceed;
  char *contpath;
  char *filename;

  newstdout=-1;
  newstderr=-1;
  phase=0;
  proceed=true;
  contpath=getcontpath(contname);
  (void) asprintf(&filename,"%s/%s.stdout",contpath,outname);
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*open the new stdout	*/
	(void) unlink(filename);
	if ((newstdout=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 1	:	/*open the new stderr	*/
	(void) free(filename);
        (void) asprintf(&filename,"%s/%s.stderr",contpath,outname);
	(void) unlink(filename);
	if ((newstderr=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
	  (void) close(newstdout);
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 2	:	/*duplicating stdout	*/
	if (dup2(newstdout,1)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stdout (error=<%s>)",
			      OPEP,strerror(errno));
	  (void) close(newstderr);
	  (void) close(newstdout);
	  phase=999;
	  }
	break;
      case 3	:	/*duplicating stderr	*/
	(void) close(newstdout);
	if (dup2(newstderr,2)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stderr (error=<%s>)",
			      OPEP,strerror(errno));
	  (void) close(newstderr);
	  phase=999;
	  }
	break;
      case 4	:	/*setting new std	*/
	if ((stdout=fdopen(1,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stdout (error=<%s>)",
			      OPEP,strerror(errno));
	  }
	if ((stderr=fdopen(2,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stderr (error=<%s>)",
			      OPEP,strerror(errno));
	  }
	(void) fclose(stdin);
	break;
      case 5	:	/*write time stamp	*/
	(void) fprintf(stdout,"stdout start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fprintf(stderr,"stderr start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fflush(stdout);
	(void) fflush(stderr);
	break;
      default	:	/*SAFE Guard		*/
	proceed=false;
        break;
      }
    phase++;
    }
  (void) free(filename);
  (void) free(contpath);
  }
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for entry on the	*/
/*	container console and forward it	*/
/*	on the Master_container console file.	*/
/*	A Management expiration is done		*/
/*	according time.				*/
/*	Status is 0 and 2, if process shutdown	*/
/*	Status is 1, if process need reboot	*/
/*						*/
/************************************************/
int cnt_mstconsole(char *contname,pid_t cntpid)

{
#define	OPCO	"unicnt.c:cnt_cmstonsole,"

#define	RLX	30	/*number of sec		*/
int count;
int relaxtime;
int status;
int nbr;
int cconsole;
int mconsole;
struct timeval relax;
fd_set reading;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[1000];
count=0;
relaxtime=5;
status=0;
nbr=0;
cconsole=0;
mconsole=0;
relax.tv_sec=relaxtime;	/*sec relax time	*/
relax.tv_usec=0;
FD_ZERO(&reading);
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"console");
phase=0;
proceed=true;
while (proceed==true) {
  count++;
  switch (phase) {
    case 0	:	/*opening the M-Console*/
      if ((mconsole=open(filename,O_CREAT|O_WRONLY|O_TRUNC,S_IRUSR|S_IWUSR))<0) {
	(void) apl_alert(0,"%s, Unable to open mconsole <%s> (error=<%s>)",
			      OPCO,filename,strerror(errno));
	phase=999;
	status=-1;
	}
      break;
    case 1	:	/*opening the C-Console*/
      (void) free(filename);
      (void) asprintf(&filename,"%s/rootfs/dev/%s",contpath,"console");
      if ((cconsole=open(filename,O_RDWR|O_ASYNC))<0) {
	(void) apl_alert(0,"%s, Unable to open cconsole <%s> (error=<%s>)",
			      OPCO,filename,strerror(errno));
  	(void) close(mconsole);
	phase=999;
	status=-1;
        }
      break;
    case 2	:		/*waiting event		*/
      relax.tv_sec=relaxtime;	/*sec relax time	*/
      relax.tv_usec=0;
      FD_ZERO(&reading);
      FD_SET(cconsole,&reading);
      switch ((nbr=select(cconsole+1,&reading,(fd_set *)0,(fd_set *)0,&relax))) {
        case	-1    :
	  switch (errno) {
	    case EINTR:	/*received a signal, lets see...*/
	      break;	/*getting out fast		*/
	    default   :
	      (void) apl_alert(0,"%s, Container <%s> unexpected select "
			      	 "event (error=<%s>)",OPCO,contname,strerror(errno));
	      break;
	    }
          break;
        case      0   :	/*timeout child lost ?	*/
	  switch (waitpid(cntpid,&status,WNOHANG)) {
	    case -1   :	/*exiting 		*/
	      (void) apl_alert(0,"Container <%s> Unexpect error on waitpid "
			      	 "(error=<%s>)",contname,strerror(errno));
	      status=-1;
	      break;
	    case 0    :	/*process still up!	*/
	      if (relaxtime<RLX)
		relaxtime++;
	      phase--;	/*lets wait for console	*/
	      break;
	    default   :	/*process exited!	*/
	      (void) apl_alert(0,"Container <%s> exited with waitpid status "
			      	  "<%d -> %s>",contname,status,strsignal(status));
	      if (WIFSIGNALED(status)==false) {
	        (void) apl_alert(2,"Container <%s> exit without signal",contname);
		status=0;
		}
	      else {
		status=WTERMSIG(status);
	        (void) apl_alert(0,"Container <%s> exited with signal "
				   "<%d -> %s>",contname,status,strsignal(status));
		}
	      break;	/*going to close phase	*/
	    }
          break;
	default	      :	/*data ready in cconsole*/
	  if ((nbr=read(cconsole,buffer,sizeof(buffer)))>0) {
	    (void) write(mconsole,buffer,nbr);
	    relaxtime=1;
	    }		/*forwarded to mconsole	*/
	  phase--;	/*stay put on phase	*/
          break;
	}
      break;
    case 3	:	/*closing console	*/
      (void) close(mconsole);
      (void) close(cconsole);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return status;
}
